(**************************************************************************)
(*                                   Ott                                  *)
(*                                                                        *)
(*        Peter Sewell, Computer Laboratory, University of Cambridge      *)
(*      Francesco Zappa Nardelli, Moscova project, INRIA Rocquencourt     *)
(*                                                                        *)
(*  Copyright 2005-2008                                                   *)
(*                                                                        *)
(*  Redistribution and use in source and binary forms, with or without    *)
(*  modification, are permitted provided that the following conditions    *)
(*  are met:                                                              *)
(*  1. Redistributions of source code must retain the above copyright     *)
(*  notice, this list of conditions and the following disclaimer.         *)
(*  2. Redistributions in binary form must reproduce the above copyright  *)
(*  notice, this list of conditions and the following disclaimer in the   *)
(*  documentation and/or other materials provided with the distribution.  *)
(*  3. The names of the authors may not be used to endorse or promote     *)
(*  products derived from this software without specific prior written    *)
(*  permission.                                                           *)
(*                                                                        *)
(*  THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS    *)
(*  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     *)
(*  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE    *)
(*  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY       *)
(*  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL    *)
(*  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE     *)
(*  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS         *)
(*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER  *)
(*  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR       *)
(*  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN   *)
(*  IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                         *)
(**************************************************************************)

open Location
open Types

exception Identity

(* command-line options *)
let colour = ref true
let source_filenames_ref = ref ([] : string list)
let isa_filename_opt = ref (None : string option)
let hol_filename_opt = ref (None : string option)
let coq_filename_opt = ref (None : string option)
let twf_filename_opt = ref (None : string option)
let tex_filename_opt = ref (None : string option)
let caml_filename_opt = ref (None : string option)
let dot_filename_opt = ref (None : string option)
let alltt_filename_opt = ref (None : string option)
let write_systemdefn_filename_opt = ref (None : string option)
let read_systemdefn_filename_opt = ref (None : string option)
let tex_name_prefix = ref "ott"
let tex_filter_filenames = ref ([] : (string * string) list)
let tex_filter_filename_srcs = ref ([] : string list)
let tex_filter_filename_dsts = ref ([] :  string list)
let isa_filter_filenames = ref ([] : (string * string) list)
let isa_filter_filename_srcs = ref ([] : string list)
let isa_filter_filename_dsts = ref ([] :  string list)
let hol_filter_filenames = ref ([] : (string * string) list)
let hol_filter_filename_srcs = ref ([] : string list)
let hol_filter_filename_dsts = ref ([] :  string list)
let coq_filter_filenames = ref ([] : (string * string) list)
let coq_filter_filename_srcs = ref ([] : string list)
let coq_filter_filename_dsts = ref ([] :  string list)
let twf_filter_filenames = ref ([] : (string * string) list)
let twf_filter_filename_srcs = ref ([] : string list)
let twf_filter_filename_dsts = ref ([] :  string list)
let caml_filter_filenames = ref ([] : (string * string) list)
let caml_filter_filename_srcs = ref ([] : string list)
let caml_filter_filename_dsts = ref ([] :  string list)
let lift_cons_prefixes = ref false
let test_parse_list = ref ([] : string list)
let sort = ref true
(* let identtest = ref true *)
let showraw = ref false
let tex_show_meta = ref true
let tex_show_categories = ref false
let tex_colour = ref true
let tex_wrap = ref true
let process_defns = ref true
let signal_parse_errors = ref false
let ascii_ugly = ref false
let show_pre_sort = ref false
let show_post_sort = ref true
let show_deps = ref false
let show_defns = ref true
let isa_syntax = ref false
let isa_primrec = ref true 
let isa_generate_lemmas = ref true
let coq_expand_debug = ref (None : string option)
let coq_avoid = ref 1
let coq_expand_lists = ref true
let merge_fragments = ref false
let picky_multiple_parses = ref false

let options = Arg.align [

(* main output stages *)
  ( "-tex", 
    Arg.String (fun s -> tex_filename_opt := Some s),
    "<filename>        Output TeX definitions" );
  ( "-coq", 
    Arg.String (fun s -> coq_filename_opt := Some s),
    "<filename>        Output Coq definitions" ); 
  ( "-hol", 
    Arg.String (fun s -> hol_filename_opt := Some s),
    "<filename>        Output HOL definitions" ); 
  ( "-isabelle", 
    Arg.String (fun s -> isa_filename_opt := Some s),
    "<filename>   Output Isabelle definitions" ); 
(*   ( "-twelf",  *)
(*     Arg.String (fun s -> twf_filename_opt := Some s), *)
(*     "<filename>        Output Twelf definitions (incomplete)" );  *)
  ( "-ocaml", 
    Arg.String (fun s -> caml_filename_opt := Some s),
    "<filename>        Output OCaml definitions" ); 
  ( "-writesys", 
    Arg.String (fun s -> write_systemdefn_filename_opt := Some s),
    "<filename>        Output system definition" ); 
  ( "-readsys", 
    Arg.String (fun s -> read_systemdefn_filename_opt := Some s),
    "<filename>        Input system definition" ); 

(* filter filenames *)
  ( "-tex_filter", 
    Arg.Tuple[Arg.String (fun s -> tex_filter_filename_srcs := s :: !tex_filter_filename_srcs);
              Arg.String (fun s -> tex_filter_filename_dsts := s :: !tex_filter_filename_dsts)],
    "<src><dst>  Files to TeX filter" ); 
  ( "-coq_filter", 
    Arg.Tuple[Arg.String (fun s -> coq_filter_filename_srcs := s :: !coq_filter_filename_srcs);
              Arg.String (fun s -> coq_filter_filename_dsts := s :: !coq_filter_filename_dsts)],
    "<src><dst>  Files to Coq filter" ); 
  ( "-hol_filter", 
    Arg.Tuple[Arg.String (fun s -> hol_filter_filename_srcs := s :: !hol_filter_filename_srcs);
              Arg.String (fun s -> hol_filter_filename_dsts := s :: !hol_filter_filename_dsts)],
    "<src><dst>  Files to HOL filter" ); 
  ( "-isa_filter", 
    Arg.Tuple[Arg.String (fun s -> isa_filter_filename_srcs := s :: !isa_filter_filename_srcs);
              Arg.String (fun s -> isa_filter_filename_dsts := s :: !isa_filter_filename_dsts)],
    "<src><dst>  Files to Isabelle filter" ); 
(*   ( "-twelf_filter",  *)
(*     Arg.Tuple[Arg.String (fun s -> twf_filter_filename_srcs := s :: !twf_filter_filename_srcs); *)
(*               Arg.String (fun s -> twf_filter_filename_dsts := s :: !twf_filter_filename_dsts)], *)
(*     "<src><dst>  Files to Twelf filter" );  *)
  ( "-ocaml_filter", 
    Arg.Tuple[Arg.String (fun s -> caml_filter_filename_srcs := s :: !caml_filter_filename_srcs);
              Arg.String (fun s -> caml_filter_filename_dsts := s :: !caml_filter_filename_dsts)],
    "<src><dst>  Files to OCaml filter" ); 
  ( "-merge", 
    Arg.Bool (fun b -> merge_fragments := b), 
    "<"^string_of_bool !merge_fragments ^">         merge grammar and definition rules" ); 

  ( "-parse", 
    Arg.String (fun s -> test_parse_list := !test_parse_list @ [s]),
    "<string>        Test parse symterm,eg \":nontermroot: term\"" ); 

(* general behaviour *)
  ( "-signal_parse_errors",
    Arg.Bool (fun b -> signal_parse_errors := b),
    "<"^string_of_bool !signal_parse_errors^">  return >0 if there are bad defns" );
  ( "-picky_multiple_parses",
    Arg.Bool (fun b -> picky_multiple_parses := b),
    "<"^string_of_bool !picky_multiple_parses^">  Picky about multiple parses" );

(* options for ascii output *)
  ( "-colour", 
    Arg.Bool (fun b -> colour := b), 
    "<"^string_of_bool !colour ^">         Use (vt220) colour for ASCII pretty print" ); 
  ( "-show_pre_sort", 
    Arg.Bool (fun b -> show_pre_sort := b), 
    "<"^string_of_bool !show_pre_sort ^"> Show ASCII pre-sort pretty print syntax" ); 
  ( "-show_post_sort", 
    Arg.Bool (fun b -> show_post_sort := b), 
    "<"^string_of_bool !show_post_sort ^"> Show ASCII post-sort pretty print syntax" ); 
  ( "-show_defns", 
    Arg.Bool (fun b -> show_defns := b), 
    "<"^string_of_bool !show_defns ^">     Show ASCII pretty print defns" ); 
(* this doesn't seem to work anymore *)
(*   ( "-lift_cons_prefixes",  *)
(*     Arg.Bool (fun b -> lift_cons_prefixes := b), *)
(*     "<"^string_of_bool !lift_cons_prefixes^"> Lift constructor prefixes back to rules in ASCII pretty prints" );  *)

(* options for tex output *)
  ( "-tex_show_meta", 
    Arg.Bool (fun b -> tex_show_meta := b),
    "<"^string_of_bool !tex_show_meta^">  Include meta prods and rules in TeX output" ); 
  ( "-tex_show_categories", 
    Arg.Bool (fun b -> tex_show_categories := b),
    "<"^string_of_bool !tex_show_categories^">  Signal production flags in TeX output" ); 
  ( "-tex_colour", 
    Arg.Bool (fun b -> tex_colour := b),
    "<"^string_of_bool !tex_colour^">     Colour parse errors in TeX output" ); 
  ( "-tex_wrap", 
    Arg.Bool (fun b -> tex_wrap := b),
    "<"^string_of_bool !tex_wrap ^">       Wrap TeX output in document pre/postamble" ); 
  ( "-tex_name_prefix", 
    Arg.String (fun s -> tex_name_prefix := s),
    "<string>  Prefix for tex commands (default \""^ !tex_name_prefix^"\")" );

(* options for isa output *)
  ( "-isabelle_primrec",
    Arg.Bool (fun b -> isa_primrec := b),
    "<"^string_of_bool !isa_primrec ^">       Use \"primrec\" instead of \"fun\" for functions" );
  ( "-isa_syntax", 
    Arg.Bool (fun b -> isa_syntax := b),
    "<"^string_of_bool !isa_syntax ^">       Use fancy syntax in Isabelle output" ); 
  ( "-isa_generate_lemmas", 
    Arg.Bool (fun b -> isa_generate_lemmas := b),
    "<"^string_of_bool !isa_syntax ^">       Lemmas for collapsed functions in Isabelle" ); 

(* options for coq output *)
  ( "-coq_avoid", 
    Arg.Int (fun i -> coq_avoid := i),
    "<"^string_of_int !coq_avoid^">          coq type-name avoidance\n                                       (0=nothing, 1=avoid, 2=secondaryify)" ); 
  ( "-coq_expand_list_types", 
    Arg.Bool (fun b -> coq_expand_lists := b),
    "<"^string_of_bool !coq_expand_lists^">  Expand list types in Coq output" ); 

(* options for debugging *)
  ( "-dot", 
    Arg.String (fun s -> dot_filename_opt := Some s),
    "<filename>        (debug) dot graph of syntax dependencies" ); 
  ( "-alltt", 
    Arg.String (fun s -> alltt_filename_opt := Some s),
    "<filename>        (debug) alltt output of (single) source file" ); 
  ( "-sort", 
    Arg.Bool (fun b -> sort := b),
    "<"^string_of_bool !sort^">           (debug) do topological sort" ); 
  (*   ( "-identtest",  *)
  (*     Arg.Bool (fun b -> identtest := b), *)
  (*     "do identity test");  *)
  ( "-process_defns", 
    Arg.Bool (fun b -> process_defns := b),
    "<"^string_of_bool !process_defns^">  (debug) process inductive reln definitions" ); 
  ( "-showraw", 
    Arg.Bool (fun b -> showraw := b),
    "<"^string_of_bool !showraw ^">       (debug) show raw grammar"); 
  ( "-ugly", 
    Arg.Bool (fun b -> ascii_ugly := b),
    "<"^string_of_bool !ascii_ugly^">          (debug) use ugly ASCII output" ); 
  ( "-coq_debug_expand", 
    Arg.String (fun s -> coq_expand_debug := Some s),
    "<filename>        (debug) Tex definitions after Coq expand" );
  ( "-no_rbcatn", 
    Arg.Bool (fun b -> Substs_pp.no_rbcatn := b),
    "<"^string_of_bool !Substs_pp.no_rbcatn^">      (debug) remove relevant bind clauses" ); 
] 

let usage_msg = 
    ("\n"
     ^ "usage: ott <options> <filename> .. <filename> \n" 
     ^ "  (use \"OCAMLRUNPARAM=p  ott ...\" to show the ocamlyacc trace)\n")

let _ = print_string ("\nOtt version "^Version.n^"   distribution of "^Version.d^"\n")

let _ = 
  Arg.parse options 
    (fun s -> source_filenames_ref := !source_filenames_ref @ [s])
    usage_msg

let _ = tex_filter_filenames := List.combine (!tex_filter_filename_srcs) (!tex_filter_filename_dsts)
let _ = hol_filter_filenames := List.combine (!hol_filter_filename_srcs) (!hol_filter_filename_dsts)
let _ = isa_filter_filenames := List.combine (!isa_filter_filename_srcs) (!isa_filter_filename_dsts)
let _ = coq_filter_filenames := List.combine (!coq_filter_filename_srcs) (!coq_filter_filename_dsts)
let _ = twf_filter_filenames := List.combine (!twf_filter_filename_srcs) (!twf_filter_filename_dsts)
let _ = caml_filter_filenames := List.combine (!caml_filter_filename_srcs) (!caml_filter_filename_dsts)
    
(* collect the proof assistant targets *)
let targets = 
  (match !hol_filename_opt with None -> [] | Some _ -> ["hol"])
  @ (match !isa_filename_opt with None -> [] | Some _ -> ["isa"])
  @ (match !coq_filename_opt with None -> [] | Some _ -> ["coq"])
  @ (match !twf_filename_opt with None -> [] | Some _ -> ["twf"])
  @ (match !caml_filename_opt with None -> [] | Some _ -> ["ocaml"]) 

(* pp modes, used both for main output and for filtered files *)
let m_ascii = Ascii { ppa_colour = !colour; 
		      ppa_lift_cons_prefixes = false; 
		      ppa_ugly= !ascii_ugly; 
		      ppa_show_deps = !show_deps; 
		      ppa_show_defns = !show_defns } 
let m_tex = Tex {ppt_colour= !tex_colour;
                 ppt_show_meta= !tex_show_meta;
                 ppt_show_categories= !tex_show_categories;
                 ppt_wrap= !tex_wrap;
                 ppt_name_prefix= !tex_name_prefix } 
let m_isa s1 = Isa { ppi_isa_primrec = !isa_primrec;
                     isa_library = Some (ref ("",[])); 
		     ppi_theory_name = s1; 
		     ppi_fancy_syntax = !isa_syntax;
		     ppi_generate_lemmas = !isa_generate_lemmas }  
let m_hol s1 = Hol { hol_library = Some (ref ("",[])); 
		     pph_theory_name = s1 }  
let m_twf s1 = Twf { twf_current_defn = ref "";
		     twf_library = Some (ref ("",[])) } 
let m_coq = Coq { coq_expand_lists = !coq_expand_lists;
		  coq_quantified_vars_from_de = ref [];
		  coq_non_local_hyp_defn = ref "";
		  coq_non_local_hyp_defn_vars = ref [];
                  coq_list_types = ref [];
                  coq_list_aux_funcs = Some (ref "");
                  coq_list_aux_defns = { defined = ref []; newly_defined = ref [] };
                  coq_list_rules_defns = ref [];
                  coq_library = Some (ref ("",[]));
		  coq_locally_nameless = ref false }  
let reset_m_coq m = 
  match m with
  | Coq co ->
      co.coq_non_local_hyp_defn := "";
      co.coq_non_local_hyp_defn_vars := [];
      co.coq_list_types := [];
      Auxl.the (co.coq_list_aux_funcs) := "";
      co.coq_list_aux_defns.defined := [];
      co.coq_list_aux_defns.newly_defined := [];
      co.coq_list_rules_defns := [];
      Auxl.the (co.coq_library) := ("",[]);
      co.coq_locally_nameless := false
  | Tex _ | Ascii _ | Isa _ | Hol _ | Twf _ | Caml _ -> Auxl.error "internal: reset_m_coq for non-coq m\n"  

let m_caml = Caml { caml_library = Some (ref ("",[])) } 

(* finally compute the set of modes used in this run of Ott -- used when non-picky about multiple parses *)

let _ =
  Global_option.is_picky := 
    ( !picky_multiple_parses,
      ( match !caml_filename_opt with None -> [] | Some _ -> [ m_caml ] )
      @ ( match !hol_filename_opt with None -> [] | Some s -> [ m_hol (Auxl.hol_filename_check s) ] ) 
      @ ( match !isa_filename_opt with None -> [] | Some s -> [ m_isa (Auxl.isa_filename_check s) ] ) 
      @ ( match !twf_filename_opt with None -> [] | Some s -> [ m_twf s ] ) 
      @ ( match !coq_filename_opt with None -> [] | Some _ -> [ m_coq ] )
      @ ( match !tex_filename_opt with None -> [] | Some _ -> [ m_tex ] ) )

(* process *)

let process source_filenames = 

  let sources = String.concat " " source_filenames in

  (match !alltt_filename_opt,source_filenames with
  | None,_ -> ()
  | Some alltt_filename,([] | _::_::_) -> 
      Auxl.error ("\nUsage: -alltt option can only be used with exactly one source file at a time\n");
  | Some alltt_filename,[source_filename] ->
    let c = open_in source_filename in
    let c' = open_out alltt_filename in
    output_string c' "\\begin{alltt}\n";
    let lexbuf = Lexing.from_channel c in
    let lexer = Grammar_lexer.my_lexer false Grammar_lexer.metalang in 
    let rec process_input () = 
      try
        let t = 
          lexer lexbuf in
        match t with 
        | Grammar_parser.EOF -> ()
        | _ -> 
            print_string (Grammar_lexer.de_lex_ascii t); flush stdout;
            output_string c' (Grammar_lexer.de_lex_tex t); flush c';
            process_input ()
      with 
        My_parse_error s->
(*      Auxl.error ("\n"^s^" in file: "^filter_filename^"\n") in*)
          Auxl.error ("\n"^s^"\n") in
    process_input ();
    output_string c' "\\end{alltt}\n";
    let _ = close_in c in
    let _ = close_out c' in ());      

  let parse_file filename = 
    let c = open_in filename in
    let lexbuf = Lexing.from_channel c in
    if List.length source_filenames <> 1 then 
      lexbuf.Lexing.lex_curr_p <- 
        { lexbuf.Lexing.lex_curr_p with Lexing.pos_fname = filename};
    let rsds = 
      try
        Grammar_parser.main (Grammar_lexer.my_lexer true Grammar_lexer.metalang) lexbuf
      with 
        My_parse_error s->
(*      Auxl.error ("\n"^s^" in file: "^filter_filename^"\n") in*)
          Auxl.error ("\n"^s^"\n") in
    let _ = close_in c in
    rsds in

  let document = List.map parse_file source_filenames in

  if !showraw then (
    let s = 
      (String.concat "\n" 
         (List.map 
            (fun document_part -> 
              String.concat "" (List.map Grammar_pp.pp_raw_syntaxdef document_part))
            document)) in
(*     let fd = open_out "test2.txt" in  *)
(*     output_string fd s ;  *)
(*     close_out fd; *)
    print_string s);

  let xd,rdcs = try
    Grammar_typecheck.check_and_disambiguate targets (!merge_fragments) document 
  with
  | Grammar_typecheck.Typecheck_error (s1,s2) ->
      Auxl.error ("\nError in checking and disambiguating syntax:\n"^s1
              ^ (if s2<>"" then " ("^s2^")" else "")
              ^ "\n")
  in
  
  if !show_pre_sort then (
    print_endline "********** AFTER CHECK AND DISAMBIGUATE  *********************\n"; 
    print_endline (Grammar_pp.pp_syntaxdefn m_ascii xd));

  let xd = Grammar_typecheck.sort_syntaxdefn xd in
  if !show_post_sort then (
    print_endline "********** AFTER SORTING  *********************\n"; 
    print_endline (Grammar_pp.pp_syntaxdefn m_ascii xd));

  (* make parser for symbolic terms *)
  let lookup = Term_parser.make_parser xd in 

  let sd = 
    if !process_defns then 
      (if !show_defns then (print_endline "********** PROCESSING DEFINITIONS *****************\n"; flush stdout);
       try
         let dcs = Defns.process_raw_defnclasss m_ascii xd lookup rdcs in
         let sd = { syntax = xd;
                    relations = dcs;
                    sources = sources} in
         sd
       with
       | Bounds.Bounds s | Defns.Rule_parse_error s | Grammar_typecheck.Typecheck_error (s,_)->
           Auxl.error ("\nError in processing definitions:\n"^s^"\n")
      )
    else
      (print_endline "********** NOT PROCESSING DEFINITIONS *************\n"; flush stdout;
       let sd = { syntax = xd;
                  relations = [];
                  sources = sources}in
       sd) in

  sd,lookup



let read_systemdefn read_systemdefn_filename =
  let fd = open_in_bin read_systemdefn_filename in
  let sd,lookup = 
    try Marshal.from_channel fd 
    with Failure s -> Auxl.error ("Cannot read dumped systemdefn\n   " ^ s ^"\n")
  in
  close_in fd;
  sd,lookup
  


let output_stage (sd,lookup) = 
  
  (** output of systemdefn *)
  ( match !write_systemdefn_filename_opt with
  | None -> ()
  | Some s ->
      let fd = open_out_bin s in
      Marshal.to_channel fd (sd,lookup) [Marshal.Closures];
      close_out fd;
      print_string ("system definition in file: " ^ s ^ "\n") );
  
  (** dot output of dependencies *)
  ( match !dot_filename_opt with
  | None -> ()
  | Some s ->
      let fd = open_out s in
      output_string fd
        (Grammar_pp.pp_dot_dep_graph pp_ascii_opts_default sd.syntax);
      close_out fd;
      print_string ("dot version in file: " ^ s ^ "\n") );


  (** tex output *)
  ( match !tex_filename_opt with
  | None -> ()
  | Some s ->
      let fd = open_out s in
      output_string fd
	(System_pp.pp_systemdefn m_tex sd lookup);
      close_out fd;
      print_string ("tex version in file: " ^ s ^ "\n") );


  (** isabelle output *)
  ( match !isa_filename_opt with
  | None -> ()
  | Some s ->
      let s1 = Auxl.isa_filename_check s in
      let fd = open_out s in
      output_string fd
	(System_pp.pp_systemdefn (m_isa s1) sd lookup);
      close_out fd;
      print_string ("Isabelle version in file: " ^ s ^ "\n") ); 


  (** hol output *)
  ( match !hol_filename_opt with
  | None -> ()
  | Some s ->
      let s1 = Auxl.hol_filename_check s in
      let fd = open_out s in
      output_string fd
	(System_pp.pp_systemdefn  (m_hol s1) sd lookup);
      close_out fd;
      print_string ("HOL version in file: " ^ s ^ "\n") ); 


  (** coq output *)
  ( match !coq_filename_opt with
  | None -> ()
  | Some s ->
      (* avoidance and secondaryification *)
      let sd = match !coq_avoid with
      | 0 -> sd
      | 1 -> Auxl.avoid_primaries_systemdefn false sd 
      | 2 -> Auxl.avoid_primaries_systemdefn true sd  
      | _ -> Auxl.error "coq type-name avoidance must be in {0,1,2}" in

      (* pp of the syntax with lists productions expanded *)
      ( match !coq_expand_debug with
      | None -> ()
      | Some s ->
          let tmp_sd = Transform.expand_lists_in_syntaxdefn m_coq sd.syntax in
          let fd = open_out (s^".tex") in
          output_string fd (System_pp.pp_systemdefn m_tex { sd with syntax = tmp_sd } lookup);
          close_out fd;
	  reset_m_coq m_coq
       );

      (* pp of the language definition *)
      let fd = open_out s in
      output_string fd
	(System_pp.pp_systemdefn m_coq sd lookup); 
      close_out fd;
      print_string ("coq version in file: " ^ s ^ "\n") );


  (** twf output *)
  ( match !twf_filename_opt with
  | None -> ()
  | Some s ->
(*       let s1 = match Auxl.string_remove_suffix s ".elf" with *)
(*       | None -> failwith ("Twelf filename must end with .elf") *)
(*       | Some s1 -> s1 in *)
      let fd = open_out s in
      output_string fd
	(System_pp.pp_systemdefn  (m_twf s) (Auxl.twelf_rename sd) lookup);
      close_out fd;
      print_string ("Twelf version in file: " ^ s ^ "\n") ); 


  (** caml output *)
  ( match !caml_filename_opt with
  | None -> ()
  | Some s ->
      let _ = match Auxl.string_remove_suffix s ".ml" with
      | None -> failwith ("OCaml filename must end with .ml")
      | Some _ -> () in
      let fd = open_out s in
      output_string fd
	(System_pp.pp_systemdefn m_caml (Auxl.caml_rename sd) lookup);
      close_out fd;
      print_string ("OCaml version in file: " ^ s ^ "\n") ); 


  (** command-line test parse *)
  (match !test_parse_list with [] -> ()|_ -> print_string "\n");
  (List.iter (function s ->
    print_string ("test parse: string \""^s^"\"\n");
    let r = Str.regexp (Str.quote ":" ^ "\\(" ^ Term_parser.ident_string ^"\\)"^ Str.quote ":" ^ "\\(.*\\)") in
    if (Str.string_match r s 0  &&
        Str.match_end () = String.length s) then
      let ntr = Str.matched_group 1 s in
      let term = Str.matched_group 4 s in
      let (ntr, concrete) = 
        if Str.string_match (Str.regexp "concrete_") ntr 0 then
          (Str.string_after ntr 9, true)
        else
          (ntr, false)
      in
      print_string ("test parse:         "^ntr^" \""^term^"\"\n");
      Term_parser.test_parse m_ascii sd.syntax ntr concrete term;
      print_string "\n"
    else
      print_string ("test parse: string "^s
                    ^" not of the required :ntr:term form\n"))
     !test_parse_list);


  (** filtering other files *)
  let filter m (src_filename,dst_filename) = 
    let fd_src = open_in src_filename in
    let fd_dst = open_out dst_filename in
    let lexbuf = Lexing.from_channel fd_src in
    lexbuf.Lexing.lex_curr_p <- 
      { lexbuf.Lexing.lex_curr_p with Lexing.pos_fname = src_filename};
    let unfiltered_document = try
      Grammar_parser.unfiltered_spec_el_list (Grammar_lexer.my_lexer true Grammar_lexer.filter) lexbuf
    with 
      Parsing.Parse_error ->
        Auxl.error ("unfiltered document "^src_filename^" cannot be parsed\n") in
    let filtered_document = Embed_pp.pp_embed_spec m sd.syntax lookup (Auxl.collapse_embed_spec_el_list unfiltered_document) in
    let _ = output_string fd_dst filtered_document in
    let _ = close_in fd_src in
    let _ = close_out fd_dst in
    ()
  in

  (List.iter  (filter m_tex) (!tex_filter_filenames));
  (List.iter  (filter m_coq) (!coq_filter_filenames));
  (List.iter  (filter (m_isa "")) (!isa_filter_filenames));
  (List.iter  (filter (m_hol "")) (!hol_filter_filenames));
  (List.iter  (filter (m_twf "")) (!twf_filter_filenames));
  (List.iter  (filter (m_caml)) (!caml_filter_filenames));
    
   
(*  let xd,rdcs = Grammar_typecheck.check_and_disambiguate targets document in  *)

  if !process_defns then begin
    let bad, msg = Defns.pp_counts sd in
    print_string msg;
    if bad && !signal_parse_errors then exit 1
  end;
  ()

    
let _ = match !source_filenames_ref, !read_systemdefn_filename_opt with
| ((_::_) as source_filenames),None -> 
    let (sd,lookup) = process source_filenames in
    output_stage (sd,lookup)
| [], Some s ->
    let (sd,lookup) = read_systemdefn s in
    output_stage (sd,lookup)
| [],None -> 
    Arg.usage options usage_msg;
    Auxl.error "\nError: must specify either source filenames or a readsys option\n"
| (_::_),Some _ -> Auxl.error "\nError: must not specify both source filenames and a readsys option\n"
