(**************************************************************************)
(*                                   Ott                                  *)
(*                                                                        *)
(*        Peter Sewell, Computer Laboratory, University of Cambridge      *)
(*      Francesco Zappa Nardelli, Moscova project, INRIA Rocquencourt     *)
(*                                                                        *)
(*  Copyright 2005-2008                                                   *)
(*                                                                        *)
(*  Redistribution and use in source and binary forms, with or without    *)
(*  modification, are permitted provided that the following conditions    *)
(*  are met:                                                              *)
(*  1. Redistributions of source code must retain the above copyright     *)
(*  notice, this list of conditions and the following disclaimer.         *)
(*  2. Redistributions in binary form must reproduce the above copyright  *)
(*  notice, this list of conditions and the following disclaimer in the   *)
(*  documentation and/or other materials provided with the distribution.  *)
(*  3. The names of the authors may not be used to endorse or promote     *)
(*  products derived from this software without specific prior written    *)
(*  permission.                                                           *)
(*                                                                        *)
(*  THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS    *)
(*  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     *)
(*  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE    *)
(*  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY       *)
(*  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL    *)
(*  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE     *)
(*  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS         *)
(*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER  *)
(*  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR       *)
(*  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN   *)
(*  IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                         *)
(**************************************************************************)

open Types;;

(* EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL *)

let pp_functions_locally_nameless m sd xd_transformed =
  let funcs_open = Ln_transform.pp_open m xd_transformed in
  let funcs_subrules = Subrules_pp.pp_subrules m xd_transformed in
  let auxfns = "" in (* Substs_pp.pp_auxfns m sd.syntax in *)
  let arities = Ln_transform.pp_arities m sd.syntax xd_transformed in
  let substs = Ln_transform.pp_substs m xd_transformed in
  let freevars = Ln_transform.pp_freevars m xd_transformed in
  let contexts = "" in (* Context_pp.pp_context m sd.syntax in *)
  let library = 
    ( match m with
    | Isa io -> fst (!(Auxl.the(io.isa_library)))
    | Hol ho -> fst (!(Auxl.the(ho.hol_library)))
    | Coq co -> fst (!(Auxl.the(co.coq_library)))
    | Twf wo -> fst (!(Auxl.the(wo.twf_library)))
    | Caml oo -> fst (!(Auxl.the(oo.caml_library)))
    | Ascii _ | Tex _ -> Auxl.error "internal: pp_functions of Tex-Ascii\n") in
  let aux_list_functions =
    ( match m with
    | Coq co -> !(Auxl.the co.coq_list_aux_funcs) 
    | _ -> "" ) in
  "(* EXPERIMENTAL *)\n"
  ^ Auxl.print_with_comment m "" "auxiliary functions on the new list types" 
    aux_list_functions
  ^ Auxl.print_with_comment m "\n" "library functions" 
      library 
  ^ Auxl.print_with_comment m "\n" "subrules"
      (Dependency.compute m xd_transformed funcs_subrules)
  ^ Auxl.print_with_comment m "\n" "arities"
      (Dependency.compute m sd.syntax arities)
  ^ Auxl.print_with_comment m "\n" "opening up abstractions"
      (Dependency.compute m xd_transformed funcs_open)
  ^ Auxl.print_with_comment m "\n" "terms are locally-closed pre-terms"
      (Ln_transform.pp_lcs m xd_transformed)  
(*   ^ Auxl.print_with_comment m "\n" "auxiliary functions" *)
(*       (Dependency.compute m sd.syntax auxfns) *)
  ^ Auxl.print_with_comment m "\n" "free variables" 
      (Dependency.compute m xd_transformed freevars) 
  ^ Auxl.print_with_comment m "\n" "substitutions" 
      (Dependency.compute m sd.syntax substs) 
(*   ^ Auxl.print_with_comment m "\n" "context application" *)
(*       (Dependency.compute m sd.syntax contexts) *)

let set_locally_nameless m =
  match m with
  | Coq co -> co.coq_locally_nameless := true
  | _ -> failwith "locally nameless implemented only for Coq"

let pp_systemdefn_core_locally_nameless m sd lookup = 
  set_locally_nameless m;
  let xd_ln_transformed = Ln_transform.ln_transform_syntaxdefn m sd.syntax in
  let relations_ln_transformed = Ln_transform.ln_transform_fun_or_reln_defnclass_list m sd.syntax sd.relations in 
  Embed_pp.pp_embeds m sd.syntax lookup sd.syntax.xd_embed_preamble
  ^ Auxl.big_line_comment m "syntax"
  ^ Grammar_pp.pp_syntaxdefn m xd_ln_transformed
  ^ "\n"
  ^ pp_functions_locally_nameless m sd xd_ln_transformed
  ^ Embed_pp.pp_embeds m sd.syntax lookup sd.syntax.xd_embed
  ^ "\n"
  ^ Defns.pp_fun_or_reln_defnclass_list 
      m xd_ln_transformed (Term_parser.make_parser xd_ln_transformed) relations_ln_transformed 
  ^ "\n"
  ^ Ln_transform.ln_infrastructure m xd_ln_transformed relations_ln_transformed
  ^ "\n"
  ^ ( match m with
    | Coq co when not co.coq_expand_lists ->
	let ip = Coq_induct.pp_induction m xd_ln_transformed in
	if ip <> "" then Auxl.big_line_comment m "induction principles" ^ ip
	else ""
    | _ -> "" ) 



(*   ^ (if sd.syntax.xd_embed_postamble <> [] then "\n" else "") *)
(*   ^ Embed_pp.pp_embeds m sd.syntax lookup sd.syntax.xd_embed_postamble *)
  ^ "\n"
(* END EXPERIMENTAL END EXPERIMENTAL END EXPERIMENTAL END EXPERIMENTAL END EXPERIMENTAL END EXPERIMENTAL *)


let pp_functions m sd =
  let funcs_subrules = Subrules_pp.pp_subrules m sd.syntax in
  let auxfns = Substs_pp.pp_auxfns m sd.syntax in
  let substs = Substs_pp.pp_substs m sd.syntax in
  let freevars = Substs_pp.pp_freevars m sd.syntax in
  let contexts = Context_pp.pp_context m sd.syntax in
  let library = 
    ( match m with
    | Isa io -> fst (!(Auxl.the(io.isa_library)))
    | Hol ho -> fst (!(Auxl.the(ho.hol_library)))
    | Coq co -> fst (!(Auxl.the(co.coq_library)))
    | Twf wo -> fst (!(Auxl.the(wo.twf_library)))
    | Caml oo -> fst (!(Auxl.the(oo.caml_library)))
    | Ascii _ | Tex _ -> Auxl.error "internal: pp_functions of Tex-Ascii\n") in
  let aux_list_functions =
    ( match m with
    | Coq co -> !(Auxl.the co.coq_list_aux_funcs) 
    | _ -> "" ) in
  Auxl.print_with_comment m "" "auxiliary functions on the new list types" 
    aux_list_functions
  ^ Auxl.print_with_comment m "\n" "library functions" 
      library 
  ^ Auxl.print_with_comment m "\n" "subrules" 
      (Dependency.compute m sd.syntax funcs_subrules)
  ^ Auxl.print_with_comment m "\n" "auxiliary functions"
      (Dependency.compute m sd.syntax auxfns)
  ^ Auxl.print_with_comment m "\n" "free variables"
      (Dependency.compute m sd.syntax freevars)
  ^ Auxl.print_with_comment m "\n" "substitutions"
      (Dependency.compute m sd.syntax substs)
  ^ Auxl.print_with_comment m "\n" "context application"
      (Dependency.compute m sd.syntax contexts)


(* core for isa/hol/coq/twf output *)
let pp_systemdefn_core m sd lookup = 
  (* FZ keeping for now locally_nameless repr separated from the main *)
  if Auxl.require_locally_nameless sd.syntax
  then pp_systemdefn_core_locally_nameless m sd lookup
  else

  let sd_expanded = Transform.expand_lists_in_syntaxdefn m sd.syntax in  (* no-op if m <> Coq *)
  Embed_pp.pp_embeds m sd.syntax lookup sd.syntax.xd_embed_preamble
  ^ Auxl.big_line_comment m "syntax"
  ^ Grammar_pp.pp_syntaxdefn m sd_expanded
  ^ "\n"
  ^ pp_functions m sd 
  ^ Embed_pp.pp_embeds m sd.syntax lookup sd.syntax.xd_embed
  ^ "\n"
  ^ Defns.pp_fun_or_reln_defnclass_list m sd.syntax lookup sd.relations
  ^ (if sd.syntax.xd_embed_postamble <> [] then "\n\n" else "\n")
  ^ Embed_pp.pp_embeds m sd.syntax lookup sd.syntax.xd_embed_postamble
  ^ "\n"
  ^ ( match m with
    | Coq co when not co.coq_expand_lists ->
	let ip = Coq_induct.pp_induction m sd_expanded in
	if ip <> "" 
	then Auxl.big_line_comment m "induction principles" ^ ip
	else ""
    | _ -> "" ) 


let pp_systemdefn m sd lookup =
  match m with
  | Ascii ao ->
      Grammar_pp.pp_syntaxdefn m sd.syntax
      ^ Defns.pp_fun_or_reln_defnclass_list m sd.syntax lookup sd.relations
  | Isa io ->
      (*  "(\* -*-SML-*- *\)\n"   (\* to get halfway decent fontlocking *\)  *)
      "(* generated by Ott "^Version.n^" from: "^sd.sources^" *)\n"
      ^ "theory "^io.ppi_theory_name^" imports Main Multiset begin\n\n" 
      ^ pp_systemdefn_core m sd lookup
      ^ "\nend\n"
  | Hol ho ->
      "(* generated by Ott "^Version.n^" from: "^sd.sources^" *)\n"
      ^ "(* to compile: Holmake "^ho.pph_theory_name^"Theory.uo   *)\n"
      ^ "(* for interactive use:\n"
      ^ "  app load [\"pred_setTheory\",\"finite_mapTheory\",\"stringTheory\",\"containerTheory\",\"ottLib\"];\n"
      ^"*)\n\n"
      ^ "open HolKernel boolLib Parse bossLib ottLib;\n"  
      ^ "infix THEN THENC |-> ## ;\n"
      ^ "local open arithmeticTheory stringTheory containerTheory pred_setTheory listTheory \n"
      ^ "  finite_mapTheory in end;\n\n"
      ^ "val _ = new_theory \""^ho.pph_theory_name^"\";\n\n" 
      ^ pp_systemdefn_core m sd lookup
(*   ^ Grammar_pp.pp_syntaxdefn m sd.syntax *)
(*   ^ Defns.pp_defnclass_list m sd.syntax sd.relations *)
(*   ^ "\n" *)

      ^ "\nval _ = export_theory ();\n"
  | Coq co ->
      (* FZ keeping for now locally_nameless repr separated from the main *)
      ( if Auxl.require_locally_nameless sd.syntax
      then
	"(* generated by Ott " ^ Version.n ^ " ***locally nameless*** from: " ^ sd.sources ^ " *)\n\n"
(*	^ "Set Implicit Arguments.\n\n" *)
	^ String.concat "\n" (List.map (fun th -> "Require Import " ^ th ^ ".") 
				( match co.coq_expand_lists with
				| true -> ["Metatheory"]
				| false -> ["Bool"; "Metatheory"; "List"; "ott_list_core"])) 

      else

	"(* generated by Ott " ^ Version.n ^ " from: " ^ sd.sources ^ " *)\n\n"
	^ String.concat "\n" (List.map (fun th -> "Require Import " ^ th ^ ".") 
				( match co.coq_expand_lists with
				| true -> ["Arith"; "Bool"; "List"]
				| false -> ["Arith"; "Bool"; "List"; "ott_list_core"])) )
      ^ "\n\n"
      ^ (pp_systemdefn_core m sd lookup) ^ "\n"
  | Twf wo ->
      "%% generated by Ott "^Version.n^" from: "^sd.sources^" \n\n"
      ^ pp_systemdefn_core m sd lookup
  | Caml oo ->
      "(* generated by Ott "^Version.n^" from: "^sd.sources^" *)\n\n"
      ^ pp_systemdefn_core m sd lookup ^ "\n"
  | Tex xo ->
      "% generated by Ott "^Version.n^" from: "^sd.sources^"\n"
      ^ (if xo.ppt_wrap then 
        "\\documentclass[11pt]{article}\n\\usepackage{amsmath,amssymb}\n\\usepackage{supertabular}\n"
        ^ 
"\\usepackage{geometry}
\\usepackage{ifthen}
\\geometry{%
a4paper,
% letterpaper,
dvips,
twoside,
left=22.5mm, right=22.5mm,
top=20mm,  bottom=30mm,
}
"
        ^ (if xo.ppt_colour then "\\usepackage{color}\n" else "")
        ^ "\\begin{document}\n"
      else
        "")
      ^ "\\newcommand{"^Grammar_pp.pp_tex_DRULE_NAME m ^"}[4][]{{\\displaystyle\\frac{\\begin{array}{l}#2\\end{array}}{#3}\\quad"^Grammar_pp.pp_tex_DRULE_NAME_NAME m^"{#4}}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_USE_DRULE_NAME m ^"}[1]{\\[#1\\]}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_PREMISE_NAME m ^"}[1]{ #1 \\\\}\n"
      ^ "\\newenvironment{"^Grammar_pp.pp_tex_DEFN_BLOCK_NAME m ^"}[3][]{ \\framebox{\\mbox{#2}} \\quad #3 \\\\[0pt]}{}\n"
      ^ "\\newenvironment{"^Grammar_pp.pp_tex_FUNDEFN_BLOCK_NAME m ^"}[3][]{ \\framebox{\\mbox{#2}} \\quad #3 \\\\[0pt]\\begin{displaymath}\\begin{array}{l}}{\\end{array}\\end{displaymath}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_FUNCLAUSE_NAME m ^"}[2]{ #1 \\equiv #2 \\\\}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_NT_NAME m ^"}[1]{\\mathit{#1}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_MV_NAME m ^"}[1]{\\mathit{#1}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_KW_NAME m ^"}[1]{\\mathbf{#1}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_COM_NAME m ^"}[1]{\\text{#1}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_DRULE_NAME_NAME m ^"}[1]{\\textsc{#1}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_COMP_LU_NAME m ^"}[5]{\\overline{#1}^{\\,#2\\in #3 #4 #5}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_COMP_U_NAME m ^"}[3]{\\overline{#1}^{\\,#2<#3}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_COMP_NAME m ^"}[2]{\\overline{#1}^{\\,#2}}\n"
      ^ "\\newcommand{\\" ^ Grammar_pp.pp_tex_NAME_PREFIX m
      ^   "grammartabular}[1]{\\begin{supertabular}{llcllllll}#1\\end{supertabular}}\n"
      ^ "\\newcommand{\\" ^ Grammar_pp.pp_tex_NAME_PREFIX m
      ^   "metavartabular}[1]{\\begin{supertabular}{ll}#1\\end{supertabular}}\n"

         (* {prod_flavour}{prodnames}{meta}{bindspec}{othercategories}{com} *)
      ^ "\\newcommand{"^Grammar_pp.pp_tex_RULEHEAD_NAME m ^"}[3]{$#1$ & & $#2$ & & & \\multicolumn{2}{l}{#3}}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_PROD_NAME m ^"}[6]{& & $#1$ & $#2$ & $#3 #4$ & $#5$ & $#6$}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_INTERRULE_NAME m ^"}{\\\\[5.0mm]}\n"
      ^ "\\newcommand{"^Grammar_pp.pp_tex_AFTERLASTRULE_NAME m ^"}{\\\\}\n"
      ^ Embed_pp.pp_embeds m sd.syntax lookup sd.syntax.xd_embed_preamble
      ^ Grammar_pp.pp_syntaxdefn m sd.syntax
      ^ Defns.pp_fun_or_reln_defnclass_list m sd.syntax lookup sd.relations
      ^ "\\newcommand{"^Grammar_pp.pp_tex_ALL_NAME m^"}{"
      ^ Grammar_pp.pp_tex_METAVARS_NAME m^"\\\\[0pt]\n"
      ^ Grammar_pp.pp_tex_RULES_NAME m^"\\\\[5.0mm]\n"
      ^ Embed_pp.pp_embeds m sd.syntax lookup sd.syntax.xd_embed
      ^ Grammar_pp.pp_tex_DEFNSS_NAME m^"\n"
      ^ Embed_pp.pp_embeds m sd.syntax lookup sd.syntax.xd_embed_postamble
      ^"}\n\n"
      ^ (if xo.ppt_wrap then
        Grammar_pp.pp_tex_ALL_NAME m^"\n"
        ^ "\\begin{verbatim}\n" 
        ^ snd (Defns.pp_counts sd)
        ^ "\\end{verbatim}\n"
        ^ "\\end{document}\n"
      else
        "")



